#include "epd.h"
#include "epd_hal.h"
#include "epd_command.h"

// ο:  https://github.com/waveshareteam/e-Paper/blob/master/Arduino/epd4in2bc/epd4in2b.cpp


static epd_uint8_t blackwhite_channel_color;
static epd_uint8_t xwhite_channel_color;

void epd_reset()
{
    epd_hal_io_write_reset(EPD_IO_LOW);
    epd_hal_delay_ms(200);
    epd_hal_io_write_reset(EPD_IO_HIGH);
    epd_hal_delay_ms(200);
}

void epd_init()
{
    epd_reset();
    epd_send_command(BOOSTER_SOFT_START);
    epd_send_data(0x17);
    epd_send_data(0x17);
    epd_send_data(0x17);
    epd_send_command(POWER_ON);
    epd_wait_until_idle();
    epd_send_command(PANEL_SETTING);
    epd_send_data(0x0F);
}

void epd_send_command(epd_uint8_t epd_command)
{
    epd_hal_io_write_dc(EPD_IO_LOW);
    epd_hal_spi_transmite(epd_command);
}

void epd_send_data(epd_uint8_t epd_data)
{
    epd_hal_io_write_dc(EPD_IO_HIGH);
    epd_hal_spi_transmite(epd_data);
}

void epd_wait_until_idle()
{
    while (epd_hal_io_read_busy() == EPD_IO_LOW)
    {
        epd_hal_delay_ms(100);
    }
}

void epd_refresh()
{
    epd_send_command(DISPLAY_REFRESH);
    epd_hal_delay_ms(100);
    epd_wait_until_idle();
}

void epd_set_clear_color(epd_uint8_t bw_ch, epd_uint8_t xw_ch)
{
    blackwhite_channel_color = bw_ch;
    xwhite_channel_color = xw_ch;
}

void epd_get_clear_color(epd_uint8_t* bw_ch, epd_uint8_t* xw_ch)
{
    *bw_ch = blackwhite_channel_color;
    *xw_ch = xwhite_channel_color;
}

void epd_clear()
{
    epd_uint8_t bw, xw;
    epd_get_clear_color(&bw, &xw);
    epd_fill_color(epd_fill_bw_ch, bw);
    epd_fill_color(epd_fill_xw_ch, xw);
    epd_refresh();
}

static void epd_select_color_channel(epd_fill_color_channel_t channel)
{
    switch (channel)
    {
        case epd_fill_bw_ch: epd_send_command(DATA_START_TRANSMISSION_1); break;
        case epd_fill_xw_ch: epd_send_command(DATA_START_TRANSMISSION_2); break;
        default: break;
    }
}

static epd_uint16_t epd_calc_buffer_size(epd_uint16_t* w, epd_uint16_t* l)
{
    static epd_uint16_t width = (EPD_WIDTH % 8 == 0) ? (EPD_WIDTH / 8) : (EPD_WIDTH / 8 + 1);
    static epd_uint16_t height = EPD_HEIGHT;
    static epd_uint16_t buffer_size = ((EPD_WIDTH % 8 == 0) ? (EPD_WIDTH / 8) : (EPD_WIDTH / 8 + 1)) * EPD_HEIGHT;
    if (w) *w = width;
    if (l) *l = height;
    return buffer_size;
}

void epd_fill_color(epd_fill_color_channel_t channel, epd_uint8_t color)
{
    epd_uint16_t index;
    epd_uint16_t buffer_size;
    
    buffer_size = epd_calc_buffer_size(NULL, NULL);
    epd_select_color_channel(channel);
    
    for (index = 0; index < buffer_size; ++index)
    {
        epd_send_data(color);
    }
    
    epd_hal_delay_ms(2);
}

void epd_fill_image(epd_fill_color_channel_t channel, const epd_uint8_t* image)
{
    epd_uint16_t index;
    epd_uint16_t buffer_size;
    
    buffer_size = epd_calc_buffer_size(NULL, NULL);
    epd_select_color_channel(channel);
    
    for (index = 0; index < buffer_size;)
    {
        epd_send_data(image[index++]);
    }
    
    epd_hal_delay_ms(2);
}

static void epd_set_partial_window(epd_rect_t rect)
{
    epd_send_command(PARTIAL_IN);
    epd_send_command(PARTIAL_WINDOW);
    epd_send_data((rect.x) >> 8);
    epd_send_data((rect.x) & 0xf8);
    epd_send_data((((rect.x) & 0xf8) + (rect.width)  - 1) >> 8);
    epd_send_data((((rect.x) & 0xf8) + (rect.width)  - 1) | 0x07);
    epd_send_data((rect.y) >> 8);        
    epd_send_data((rect.y) & 0xff);
    epd_send_data(((rect.y) + (rect.height) - 1) >> 8);        
    epd_send_data(((rect.y) + (rect.height) - 1) & 0xff);
    epd_send_data(0x01);
    epd_hal_delay_ms(2);
}

static void reset_partial_window()
{
    epd_hal_delay_ms(2);
    epd_send_command(PARTIAL_OUT);  
}

void epd_fill_color_with_rect(epd_fill_color_channel_t channel, epd_uint8_t color, epd_rect_t rect)
{
    epd_uint16_t index;
    epd_uint16_t buffer_size;
    
    epd_set_partial_window(rect);
    
    epd_select_color_channel(channel);
    
    buffer_size = rect.width / 8 * rect.height;
    
    for (index = 0; index < buffer_size; ++index)
    {
        epd_send_data(color);
    }
    
    reset_partial_window();
}

void epd_fill_image_with_rect(epd_fill_color_channel_t channel, const epd_uint8_t* image, epd_rect_t rect)
{
    epd_uint16_t index;
    epd_uint16_t buffer_size;
    
    epd_set_partial_window(rect);
    
    epd_select_color_channel(channel);
    
    buffer_size = rect.width / 8 * rect.height;
    
    for (index = 0; index < buffer_size;)
    {
        epd_send_data(image[index++]);
    }
    
    reset_partial_window();
}

void epd_sleep()
{
    epd_send_command(VCOM_AND_DATA_INTERVAL_SETTING);
    epd_send_data(0xF7);
    epd_send_command(POWER_OFF);
    epd_wait_until_idle();
    epd_send_command(DEEP_SLEEP);
    epd_send_data(0xA5);
}
